#include <iostream>
#include <sstream>

#include <glibmm.h>
#include <sqlite3_bmp.h>

#include <boost/tuple/tuple.hpp>
#include <boost/variant.hpp>
#include <boost/format.hpp>

#include "debug.hpp"
#include "database.hpp"

namespace
{

  const char* sql_valuenames[] =
  {
      "INTEGER",    //SQLite doesn't know a boolean type so we have to convert hence and forth
      "INTEGER",
      "REAL",
      "TEXT"
  };

  static boost::format format_int  ("%d");
  static boost::format format_bool ("%d");
  static boost::format format_real ("%f");

  int statement_prepare (sqlite3_stmt **pStmt, const Glib::ustring& sql, sqlite3 *sqlite)
  {
      int status;
      const char *tail;
      status = sqlite3_prepare (sqlite,
                                sql.c_str(), 
                                strlen (sql.c_str()),
                                pStmt,
                                &tail);      
  
      if (status)
    {
      g_warning ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.c_str ());
      sqlite3_reset (*pStmt);
    };

      return status;
  }

  void get_row_from_statement (sqlite3_stmt *pStmt, Bmp::DB::DataRow& row, Bmp::DB::ValueMap const& map)
  {
      using namespace Bmp::DB;

      unsigned int counter = 0;
      for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
      {
        const PairValueType &pair = (*iter);
        switch (pair.second)
          {
            case VALUE_TYPE_BOOL:
              row.insert (std::make_pair (pair.first, bool(sqlite3_column_int (pStmt, counter++)))); break;

            case VALUE_TYPE_INT:
              row.insert (std::make_pair (pair.first, int(sqlite3_column_int (pStmt, counter++)))); break;

            case VALUE_TYPE_REAL:
              row.insert (std::make_pair (pair.first, double(sqlite3_column_double (pStmt, counter++)))); break;

            case VALUE_TYPE_STRING:
              {
                const char* value = (const char*)sqlite3_column_text (pStmt, counter++);        
                if (value)
                  {
                    row.insert (std::make_pair (pair.first, Glib::ustring(value)));
                    break;
                  }
              }
            default: break;
          }
      }
  }

  void
  print_compiled_statement (sqlite3_stmt *stmt)
  {
    int count = sqlite3_bind_parameter_count (stmt);
    for (int n = 0; n < count; ++n)
    {
      g_message ("Parameter %d: %s", n, sqlite3_bind_parameter_name (stmt, n));
    }
  }

  void
  append_attributes (Glib::ustring &sql, const Bmp::DB::AttributeList& attributes)
  {
      using namespace Bmp;
      using namespace Bmp::DB;
      AttributeList::const_iterator a_last (attributes.end ());
      a_last--;
      for (AttributeList::const_iterator a_iter  = attributes.begin  (),
                     a_end   = attributes.end    ();
                     a_iter != a_end ; ++a_iter)
      {
        Bmp::DB::Attribute const& Attr      = (*a_iter);
        const bool exact            = boost::get<0>(Attr);
        std::string const& attribute_name   = boost::get<1>(Attr);
        Variant const& value            = boost::get<2>(Attr);

        if (!exact)
          {
            if (value.which() == VALUE_TYPE_STRING)
              {
                char *esc = sqlite3_mprintf (" %s LIKE '%%%q%%' ", 
                         attribute_name.c_str(),
                         boost::get<Glib::ustring>(value).c_str()); 
                sql.append (esc);
                g_free (esc);
              }
            else
              {
                sql.append (attribute_name).append (" LIKE '%").append ((format_int % boost::get<int>(value)).str()).append ("%' ");
              }
          } 
        else
          {
            if (value.which() == VALUE_TYPE_STRING)
            {
              char *esc = sqlite3_mprintf (" %s='%q' ", 
                         attribute_name.c_str(),
                         boost::get<Glib::ustring>(value).c_str()); 
              sql.append (esc);
              g_free (esc);
            }
          else
            {
              sql.append (attribute_name).append("='").append ((format_int % boost::get<int>(value)).str()).append ("' ");
            }
          }
      if (a_iter != a_last) sql.append (" AND ");
   }
  }
};

namespace Bmp
{
      namespace DB
      {
      void
      DB::sqlite_trace   (void*         data,
                  const char*   sqltext)
      {
        g_message (G_STRLOC ": %s", sqltext);
      }

      unsigned int
      DB::sqlite_exec_simple (const char* sql) 
      {
        int            status = 0;
        sqlite3_stmt  *pStmt  = 0;
        const char    *tail   = 0;
        unsigned int   rows   = 0;

        status = sqlite3_prepare (sqlite,
                                  sql,    
                                  strlen (sql),
                                  &pStmt,
                                  &tail);      

        if (status != SQLITE_OK)
          {
            sqlite3_reset (pStmt);
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }

        for (status = -1; status != SQLITE_DONE; )
          {
            status = sqlite3_step (pStmt);
            if (status == SQLITE_ROW) { rows++; continue; }
            if (status == SQLITE_BUSY) continue;
            if (status == SQLITE_DONE) break;
            if (status == SQLITE_ERROR || status == SQLITE_MISUSE)
              {
                sqlite3_reset (pStmt);
                throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
              }
          }
        sqlite3_finalize (pStmt);
        return rows;
      }

      void  
      DB::run_query (const Glib::ustring& query)
      {
        sqlite_exec_simple (query.c_str());
      }

      DB::DB (std::string name, std::string path, bool truncate) : name (name), path (path)
      {
          std::string db_filename (Glib::build_filename (path, name) + ".mlib");        

          if (Glib::file_test (db_filename, Glib::FILE_TEST_EXISTS) && truncate)
          unlink (db_filename.c_str());

          if (sqlite3_open (db_filename.c_str(), &sqlite))
            {
              throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), db_filename);
            }
#ifdef SQLITE_TRACE
          sqlite3_trace (sqlite, DB::sqlite_trace, this);
#endif
      };

      DB::~DB ()
      {
          sqlite3_close (sqlite);
      };

      void
      DB::create_attribute_view     (const std::string&   name,
                     const std::string&   view_name,
                     const std::string&   attribute,
                     const ValueMap&      map)  
      {
        mapmap.insert (std::make_pair (name, map));

        if (table_exists (view_name)) return;

        Glib::ustring sql;
        sql.append ("CREATE VIEW ").append(name).append(".").append(view_name).append(" AS SELECT DISTINCT ").append(attribute).
                        append(" FROM " ).append(name).append(";");
        sqlite_exec_simple (sql.c_str());
      }

      void
      DB::create_view_manual    (const std::string&  name,
                                 const std::string&  view_name,
                                 const std::string&  select,
                                 const ValueMap&     map)
      {
        mapmap.insert (std::make_pair (view_name, map));

        if (table_exists (view_name)) return;

        Glib::ustring sql;
        sql.append ("CREATE VIEW ").append(name).append(".").append(view_name).append(" AS ").append(select).append(";");
        sqlite_exec_simple (sql.c_str());
      }

      void
      DB::add  (const std::string&       name,
                const Bmp::DB::DataRow&  row)
      {
        if (mapmap.find (name) == mapmap.end ())
        {
          g_error ("No value-map for table '%s'", name.c_str());
        }

        const ValueMap& map = (*(mapmap.find(name))).second;

        Glib::ustring sql_keys,
                      sql_values;

        bool          had_value; 

        had_value  =  false;

        sql_keys.append ("(");
        sql_values.append ("("); 

        for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
        {
          const PairValueType &pair = (*iter);
          const Bmp::DB::DataRow::const_iterator data_iter = row.find (pair.first);

          if (data_iter != row. end ())
            {
              const Variant &value = (*data_iter).second;

              if (had_value)
                {
                  sql_keys.append(" , ");
                  sql_values.append(" , ");
                }

                //add key
                sql_keys.append (pair.first);

                //add its value
                sql_values.append ("'");
                switch (pair.second)
                {
                  case VALUE_TYPE_STRING:
                    {
                      char  *value_escaped = sqlite3_mprintf ("%q", boost::get<Glib::ustring>(value).c_str()); 
                      sql_values.append (value_escaped);
                      g_free (value_escaped);
                      break;
                    }
  
                  case VALUE_TYPE_INT:
                    sql_values.append ((format_int % boost::get<int>(value)).str()); break;

                  case VALUE_TYPE_BOOL:
                    sql_values.append ((format_bool % boost::get<bool>(value)).str()); break;

                  case VALUE_TYPE_REAL:
                    sql_values.append ((format_real % boost::get<double>(value)).str()); break;

                }
              sql_values.append ("'");
              had_value = true;
            }
        }
        sql_keys.append (")");
        sql_values.append (")");

        Glib::ustring statement;
        statement.append("INSERT INTO ").append(name).append(" ").append(sql_keys).append(" VALUES ").append(sql_values).append(";");

        debug ("sql", "Statement: %s", statement.c_str()); 

        sqlite_exec_simple (statement.c_str());
      }

      void
      DB::del   (const std::string& name,
             const AttributeList&   attributes)
      {
        Glib::ustring sql;
        sql.append ("DELETE FROM ").append(name).append(" WHERE ");
        append_attributes (sql, attributes);
        sql.append (";");
        sqlite_exec_simple (sql.c_str());
      }

      void 
      DB::project     (const std::string&       name,
               const std::string&       project, 
               VDataRow&            vector,
               const AttributeList&     attributes)
      {
        if (mapmap.find (name) == mapmap.end ())
          {
            g_error ("No value-map for table '%s'", name.c_str());
          }

        const ValueMap& map = (*(mapmap.find(name))).second;
            ValueMap::const_iterator vm_last (map.end());   
        vm_last--;

        Glib::ustring sql;
        if (!attributes.empty ())
          {
            sql.append ("SELECT DISTINCT ").append(project).append(" FROM ").append(name).append(" WHERE ");
            append_attributes (sql, attributes);
            sql.append (";");
          }
        else
          {
            sql.append ("SELECT DISTINCT ").append(project).append(" FROM ").append(name).append(";");
          }

        sqlite3_stmt  *pStmt;
        if (statement_prepare (&pStmt, sql, sqlite))
          {
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }

        for (int status = -1 ; status != SQLITE_DONE ; )
          {
            status = sqlite3_step (pStmt);
            switch (status)
              {
                case SQLITE_BUSY: continue;
                case SQLITE_DONE: break;
                case SQLITE_ROW:            
                {
                  Bmp::DB::DataRow row;
                  Bmp::DB::ValueType type = (*(map.find (project))).second;

                  switch (type)
                  {
                    case VALUE_TYPE_BOOL:
                      row.insert (std::make_pair (project, bool(sqlite3_column_int (pStmt, 0)))); break;

                    case VALUE_TYPE_INT:
                      row.insert (std::make_pair (project, int(sqlite3_column_int (pStmt, 0)))); break;

                    case VALUE_TYPE_REAL:
                      row.insert (std::make_pair (project, double(sqlite3_column_double (pStmt, 0)))); break;

                    case VALUE_TYPE_STRING:
                    {
                      const char* value = (const char*)sqlite3_column_text (pStmt, 0); 
                      if (value)
                        row.insert (std::make_pair (project, Glib::ustring(value)));
                      break;
                    }
                    default: break;
                  }

                  vector.push_back (row);
                  break;
                }
                case SQLITE_ERROR: sqlite3_reset (pStmt); return; 
                case SQLITE_MISUSE: sqlite3_reset (pStmt); return; 
              } 
          } 
        sqlite3_finalize (pStmt);
      }

      void
      DB::txn_begin       ()
      {
        Glib::ustring sql;
        sql.append ("BEGIN;");
        sqlite_exec_simple (sql.c_str());
      }

      void
      DB::txn_commit      ()
      {
        Glib::ustring sql;
        sql.append("COMMIT;");
        sqlite_exec_simple (sql.c_str());
      }

      void
      DB::txn_rollback    ()
    
      {
        Glib::ustring sql;
        sql.append("ROLLBACK;");
        sqlite_exec_simple (sql.c_str());
      }

      void 
      DB::set     (const std::string&       name,
               const std::string&       pkey,
               const std::string&       pkey_value,
               const AttributeListSet&  attributes)
      {
        Glib::ustring sql;
        sql.append ("UPDATE ").append(name).append(" SET ");
        AttributeListSet::const_iterator a_last = attributes.end ();
        a_last--;
        for (AttributeListSet::const_iterator a_iter  = attributes.begin (),
                              a_end   = attributes.end ();
                              a_iter != a_end;
                              ++a_iter) 
        {
            Bmp::DB::AttributeSet const& Attr   = (*a_iter);
            std::string const& attribute_name   = boost::get<0>(Attr);
            Variant const& value        = boost::get<1>(Attr);

            if (value.which() == VALUE_TYPE_STRING)
              {
                char *esc = sqlite3_mprintf ("%s='%q' ", attribute_name.c_str(), boost::get<Glib::ustring>(value).c_str()); 
                sql.append (esc);
                free (esc);
              }
            else
              {
                sql.append (attribute_name).append("='").append((format_int % boost::get<int>(value)).str()).append("' ");
              }

            if (a_iter != a_last) sql.append (" , ");
        }

        char *esc = sqlite3_mprintf (" WHERE %s='%q';", pkey.c_str(), pkey_value.c_str()); 
        sql.append (esc);
        free (esc);
        sqlite_exec_simple (sql.c_str());
      }

      void 
      DB::get_by_rowid (const std::string&    name,
                sqlite_int64      rowid,
                DataRow&          row)
      {
        if (mapmap.find (name) == mapmap.end ())
          {
            g_error ("No value-map for table '%s'", name.c_str());
          }

        const ValueMap& map = (*(mapmap.find(name))).second;
            ValueMap::const_iterator vm_last (map.end());   
        vm_last--;

        Glib::ustring sql;
        sql.append ("SELECT ");

        for (ValueMap::const_iterator vm_iter = map.begin     (),
                          vm_end  = map.end   ();
                          vm_iter != vm_end ; ++vm_iter)
          {
            const PairValueType &pair = (*vm_iter);
            sql.append (pair.first);
            if (vm_iter != vm_last) sql.append (",");
          }

        static boost::format lld ("%lld");
        sql.append (" FROM ").append(name).append(" WHERE rowid='").append((lld % rowid).str()).append("';"); 
        sqlite3_stmt  *pStmt;
        if (statement_prepare (&pStmt, sql, sqlite))
          { 
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }

        for (int status = -1 ; status != SQLITE_DONE ; )
          {
            status = sqlite3_step (pStmt);
            switch (status)
              {
                case SQLITE_BUSY: continue;
                case SQLITE_DONE: break;
                case SQLITE_ROW:            
                  {
                    get_row_from_statement (pStmt, row, map);
                    sqlite3_finalize (pStmt);
                    return;
                  }
                case SQLITE_ERROR: sqlite3_reset (pStmt); return;
                case SQLITE_MISUSE: sqlite3_reset (pStmt); return;
              } 
          } 
        sqlite3_finalize (pStmt);
      }

      void 
      DB::get         (const std::string&     name,
               const AttributeList&   attributes,
               VDataRow&          vector,
               const std::string&     suffix)
      {
        if (mapmap.find (name) == mapmap.end ())
        {
          g_error ("No value-map for table '%s'", name.c_str());
        }

        ValueMap const& map = (*(mapmap.find(name))).second;
            ValueMap::const_iterator vm_last (map.end());   
        vm_last--;

        Glib::ustring sql;
        sql.append ("SELECT ");
        for (ValueMap::const_iterator vm_iter = map.begin     (),
                          vm_end  = map.end   ();
                          vm_iter != vm_end ; ++vm_iter)
          {
            PairValueType const& pair = (*vm_iter);
            sql.append (pair.first);
            if (vm_iter != vm_last) sql.append (",");
          }

        if (!attributes.empty ())
          {
            sql.append (" FROM ").append(name).append(" WHERE ");
            append_attributes (sql, attributes);
            if (suffix.empty())
              {
                sql.append (";");
              }  
            else
              {
                sql.append (" ").append(suffix).append(" ").append(";");
              }
          }
        else
          {
            sql.append ("FROM ").append(name);
            if (suffix.empty())
              sql.append (";");
            else
              sql.append (" ").append(suffix).append(";");
          }

        sqlite3_stmt  *pStmt;
        if (statement_prepare (&pStmt, sql, sqlite))
          {
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }
 
        for (int status = -1 ; status != SQLITE_DONE ; )
          {
            status = sqlite3_step (pStmt);
            switch (status)
              {
                case SQLITE_BUSY: continue;
                case SQLITE_DONE: break;
                case SQLITE_ROW:            
                  {
                    Bmp::DB::DataRow row;
                    get_row_from_statement (pStmt, row, map);
                    vector.push_back (row);
                    break;
                  }
                case SQLITE_MISUSE: sqlite3_reset (pStmt); return;
                case SQLITE_ERROR: sqlite3_reset (pStmt); return;
              } 
          } 
          sqlite3_finalize (pStmt);
      }

      void
      DB::create_table (const std::string&    name,
                const std::string&    pkey,
                const ValueMap&   map)
      {
        mapmap.insert (std::make_pair (name, map));
          
        if (table_exists (name)) return;

        Glib::ustring sql;
        sql.append ("CREATE TABLE IF NOT EXISTS ").append(name).append(" (");

        ValueMap::const_iterator last;
        last = map.end ();
        last--;

        for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
          {
            PairValueType const& pair = (*iter);
            if (!pkey.compare (pair.first))
              {
                sql.append ("'").append(pair.first).append("' ").append(sql_valuenames[pair.second]).append(" NOT NULL PRIMARY KEY");
              }
            else
              {
                sql.append ("'").append(pair.first).append("' ").append(sql_valuenames[pair.second]).append(" ");
              }
            if (iter != last) sql.append (",");
          }
        sql.append (");");
        sqlite_exec_simple (sql.c_str());
      }

      void
      DB::drop_table      (const std::string& name)
      {
        Glib::ustring sql;
        sql.append ("DROP TABLE ").append(name).append(";");
        sqlite_exec_simple (sql.c_str());
      }

      bool
      DB::table_exists    (const std::string& name)
      {
        Glib::ustring sql; 
        sql.append ("SELECT name FROM sqlite_master WHERE name='").append(name).append("';");
        unsigned int nrow = sqlite_exec_simple (sql.c_str()); 
        return (nrow != 0); 
      }

      bool
      DB::pkey_exists     (const std::string&   name, 
                           const std::string&   pkey_name,
                           const Glib::ustring& pkey_value)
      {
        Glib::ustring sql; 
        char *escaped = sqlite3_mprintf("%q", pkey_value.c_str());
        sql.append ("SELECT ").append (pkey_name).append(" FROM ").append(name).append(" WHERE ").append(pkey_name).append("='").append(Glib::ustring(escaped)).append("';");
        free (escaped);
        unsigned int nrow = sqlite_exec_simple (sql.c_str()); 
        return (nrow != 0); 
      }

      void
      DB::pkey_remove     (const std::string&   name, 
                           const std::string&   pkey_name,
                           const Glib::ustring& pkey_value)
      {
        Glib::ustring sql; 
        char *escaped = sqlite3_mprintf("%q", pkey_value.c_str());
        sql.append ("DELETE FROM ").append(name).append(" WHERE ").append(pkey_name).append("='").append(Glib::ustring(escaped)).append("';");
        free (escaped);
        sqlite_exec_simple (sql.c_str()); 
      }

  } // DB

}; // Bmp
