#ifndef BMP_LIBRARY_CLASS_HPP
#define BMP_LIBRARY_CLASS_HPP

#include <set>
#include <gtkmm.h>
#include <glibmm/i18n.h>
#include <boost/optional.hpp>

#ifndef BMP_PLUGIN_BUILD
#  include <taglib/tag.h>
#endif

#include "database.hpp"
#include "util.hpp"

namespace Bmp
{

  class LibraryUI;

  namespace Library
  {
    class Exception : public std::exception
    {
      public:

        enum ExceptionID
        {
          NO_METADATA,
          NO_ASIN,
          NO_RELEASE,
          NO_FILE,
          FILE_INVALID,
          FILE_NON_READABLE,
          FILE_NON_WRITABLE,
          CONVERSION_ERROR,
          UNABLE_TO_MODIFY,

          N_IDS
        }; 
        
        enum ExceptionItemType
        {
          ITEM_ITEM,
          ITEM_URI,
        };

        Exception (const std::string& what, ExceptionID id, const std::string& item)
            : what_ (what), id_ (id), item_ (item), type_ (ITEM_ITEM)
        {
          g_log (BMP_LIBRARY_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "what (\"%s\") id (\"%s\"), item (\"%s\")", what.c_str(), _message_types[id], item_.c_str());
        }

        Exception (const std::string& what, ExceptionID id, const Glib::ustring& uri)
            : what_ (what), id_ (id), uri_ (uri), type_ (ITEM_URI)
        {
          g_log (BMP_LIBRARY_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "what (\"%s\") id (\"%s\"), uri (\"%s\")", what.c_str(), _message_types[id], uri_.c_str());
        }

        virtual ~Exception () throw()
        {}

        virtual const char *
        what () const throw()
        {
          return what_.c_str();
        }

        const char * item () const
        {
          if (type_ == ITEM_ITEM)
            return item_.c_str();
          else
            return uri_.c_str();
        }

        ExceptionID id () const
        {
          return id_;
        }

        ExceptionItemType type () const
        {
          return type_;
        }

      private:
        
        static char* const _message_types[];

        std::string             what_;
        ExceptionID             id_;
        std::string             item_;
        Glib::ustring           uri_;
        const ExceptionItemType type_;
    };
        
    enum Datum
    {
        DATUM_LOCATION,
        DATUM_ARTIST,
        DATUM_ALBUM,
        DATUM_TITLE,
        DATUM_TRACK,
        DATUM_TIME,
        DATUM_GENRE,
        DATUM_COMMENT,
        DATUM_RATING,
        DATUM_DATE,
        DATUM_MTIME,
        DATUM_BITRATE,
        DATUM_SAMPLERATE,
        DATUM_COUNT,
        DATUM_ASIN,
#ifdef HAVE_HAL
        DATUM_HAL_VOLUME_UDI,
        DATUM_HAL_DEVICE_UDI,
        DATUM_HAL_PATH,       // Path of the file relative to the volume's mount path
#endif
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_TRACK_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ARTIST_ID,

        DATUM_LAST
    };

    struct DatumDefine
    {
          const char    *title;
          const char    *title_plural;
          const char    *description;
          const char    *id;
          DB::ValueType  type;
          bool           filter;
    };

    struct Track
    {
      boost::optional <Glib::ustring>   location;
      boost::optional <Glib::ustring>   artist;
      boost::optional <Glib::ustring>   album;
      boost::optional <Glib::ustring>   title;
      boost::optional <Glib::ustring>   genre;
      boost::optional <Glib::ustring>   comment;
#ifdef HAVE_HAL
      boost::optional <Glib::ustring>   volume_udi;
      boost::optional <Glib::ustring>   device_udi;
      boost::optional <Glib::ustring>   hal_path;
#endif
      boost::optional <Glib::ustring>   asin;
      boost::optional <Glib::ustring>   mb_artistid;
      boost::optional <Glib::ustring>   mb_albumid;
      boost::optional <Glib::ustring>   mb_album_artistid;
      boost::optional <Glib::ustring>   mb_trackid;

      boost::optional <int>     tracknumber;
      boost::optional <int>     duration;
      boost::optional <int>     rating;
      boost::optional <int>     date;
      boost::optional <int>     mtime;
      boost::optional <int>     bitrate;
      boost::optional <int>     samplerate;
      boost::optional <int>     count;

      Track (const Bmp::DB::DataRow& row);
      Track () {};
      ~Track ();
    };

    struct UpdateTrack
    {
      boost::optional <Glib::ustring>   location;
      boost::optional <Glib::ustring>   artist;
      boost::optional <Glib::ustring>   album;
      boost::optional <Glib::ustring>   title;
      boost::optional <Glib::ustring>   genre;
      boost::optional <Glib::ustring>   comment;
      boost::optional <Glib::ustring>   asin;
      boost::optional <Glib::ustring>   mb_artistid;
      boost::optional <Glib::ustring>   mb_albumid;
      boost::optional <Glib::ustring>   mb_album_artistid;
      boost::optional <Glib::ustring>   mb_trackid;

      boost::optional <Glib::ustring>   location_new;

      boost::optional <int>     tracknumber;
      boost::optional <int>     date;
      boost::optional <int>     rating;
      boost::optional <int>     mtime;

      UpdateTrack& operator=(const Track& track)
      {
        location           = track.location;
        artist             = track.artist;
        album              = track.album;
        title              = track.title;
        genre              = track.genre;
        comment            = track.comment;

        asin               = track.asin;

        mb_artistid        = track.mb_artistid;
        mb_albumid         = track.mb_albumid;
        mb_album_artistid  = track.mb_album_artistid;
        mb_trackid         = track.mb_trackid;

        tracknumber        = track.tracknumber;
        date               = track.date;
        rating             = track.rating;
        mtime              = track.mtime;
      
        return *this;
      }
    };

    typedef std::vector < UpdateTrack  > UpdateTracks;

    class Library
      {
        public:

          typedef sigc::signal <void, int> ModifyStart;
          typedef sigc::signal <void>      ModifyStep;
          typedef sigc::signal <void>      ModifyEnd;
          typedef sigc::signal <void>      SignalProcessing;
          typedef sigc::signal <void, Glib::ustring, Bmp::Library::UpdateTrack > SignalRowUpdated; // returns: void, args: pkey, datarow

          SignalProcessing& signal_processing_start ();
          SignalProcessing& signal_processing_end   ();
          SignalRowUpdated& signal_row_updated      ();

          ModifyStart& signal_modify_start ();
          ModifyStep&  signal_modify_step  ();
          ModifyEnd&   signal_modify_end   ();

          enum
          {
              MB_FAILURE,
              MB_OK,
              MB_OFFLINE
          };

          typedef std::set < Glib::ustring > Compilations;

          Library ();
          ~Library ();

          void
          query     (const DB::AttributeList& attrs,
                     DB::VDataRow&            vector);

          void
          project   (const std::string&       project,
                     DB::VDataRow&            vector,
                     const DB::AttributeList& attrs = DB::AttributeList ());

          void
          increase_play_count   (const Glib::ustring& uri);

          Glib::ustring
          get_release_for_album (const Glib::ustring& album); // FIXME: The only qualifier here is the album name, not artist+album
          void
          set_asin_for_album    (const Glib::ustring& artist,
                                 const Glib::ustring& album,
                                 const Glib::ustring& asin);
          Glib::ustring
          get_asin_for_album    (const Glib::ustring& album); // FIXME: The only qualifier here is the album name, not artist+album


          void
          update_compilations ();
          bool
          is_compilation        (const Glib::ustring& album);
       


          void
          add_track    (const Glib::ustring &uri);

          bool
          add_audiocd   (const std::string &cdindex_id);

          bool
          modify_tracks (const UpdateTracks& tracks, bool retag = true);

          void
          del           (const DB::AttributeList &attrs);


          void    
          get_metadata  (const Glib::ustring &pkey,
                         DB::DataRow &row);

          GHashTable*
          get_hashtable (const DB::DataRow& row);

          const std::string
          get_metadatum_id      (Datum metadatum) const;

          const DatumDefine 
          get_metadatum_define  (Datum metadatum) const;

          Glib::ustring
          get_titlestring       (const Glib::ustring& uri, const std::string& format_);

    protected:

          friend class Bmp::LibraryUI;

#ifndef BMP_PLUGIN_BUILD
          void
          metadata_set_taglib_common (const UpdateTrack& track, TagLib::Tag* tag);
#endif
          void 
          metadata_set_taglib (const Glib::ustring &uri, const UpdateTrack& track);
          void 
          metadata_get_taglib (const Glib::ustring &uri, DB::DataRow &row);
          void
          metadata_get_gst    (const Glib::ustring &uri, DB::DataRow &row);
          void
          recreate_db ();
          
    private:

          SignalProcessing signal_processing_start_;
          SignalProcessing signal_processing_end_;
          SignalRowUpdated signal_row_updated_;

          ModifyStart  signal_modify_start_;
          ModifyStep   signal_modify_step_;
          ModifyEnd    signal_modify_end_;

          DB::ValueMap  map;
          DB::DB       *db;

          DB::ValueMap  map_compilations;
          DB::ValueMap  map_asin;
          DB::ValueMap  map_releases;
          Compilations  compilations;

    }; // class Library

  } // Library namespace

} // Bmp namespace

#endif // BMP_LIBRARY_CLASS_HPP
